# m68k-coff-as -m68000 --register-prefix-optional --bitwise-or -o sega_gcc.o sega_gcc.s
# m68k-coff-as -m68000 --register-prefix-optional --bitwise-or -o charset.o charset.s
# m68k-coff-as -m68000 --register-prefix-optional --bitwise-or -o vgmplay.o vgmplay.s
# m68k-coff-ld -Tmd.ld sega_gcc.o charset.o vgmplay.o 
# m68k-coff-objcopy -O binary a.out vgmplay.bin
# copy /B /Y vgmplay.bin + mysong.vgm mysong.bin


#################################################
#                                               #
#                 RAM DATA                      #
#                                               #
#################################################

.bss
.globl htimer
.globl vtimer
.globl rand_num
htimer:		.long 0
vtimer:		.long 0
rand_num:	.long 0
temp:		.long 0
counter1:	.long 0
counter2:	.long 0
loopPoint:	.long 0

.text
.globl main


.include "genvdp.inc"


.macro PUTSXY immstr x y
	lea	1$,a1
	move.l	\x,d0
	move.l	\y,d1
	jsr	puts
	bra 2$
1$:	.ascii "\immstr"
	dc.b 0
	.even
2$:
.endm


##################################################
#                                                #
#               MAIN PROGRAM                     #
#                                                #
##################################################
 
main:
	
	/* Initialize VDP */
	jsr 		init_gfx

	/* Wait a few frames */	
	jsr		wait_vsync
	jsr		wait_vsync
	jsr		wait_vsync
	
	/* Load font tiles */
	move.l		#0x0000,a3
	move.l 		#charset,a4
	move.l		#64,d4
	jsr		load_tiles
	move.l		#0x0800,a3
	move.l 		#(charset+0x0c00),a4
	move.l		#32,d4
	jsr		load_tiles

	/* Set the first two colors of palette 0 */
	move.l		#0x0000,a3
	move.l 		#palette,a4
	move.l		#2,d4
	jsr		load_colors

	jsr		wait_vsync

	/* Clear map A */
	move.l		#0xe000,a0
	moveq.l		#0,d0
	jsr		clear_map
	
	jsr		wait_vsync

	/* Clear map B */
	move.l		#0xc000,a0
	moveq.l		#0,d0
	jsr		clear_map
	
	jsr		wait_vsync

	/* Print some text on Map A */
	PUTSXY		"Genesis VGM Player",#7,#10
	PUTSXY		"Mic 2008",#12,#12

	
	/* Clear sprite data */
	move.l 		#GFXCNTL,a3
	VRAM_ADDR 	d0,0xfc00
	move.l 		d0,(a3)
	move.l 		#GFXDATA,a3
	moveq.l		#0,d0	
	move.l		#19,d1
clear_sprites:
	move.w		d0,(a5)+
	dbra 		d1,clear_sprites


	move.l 		#music_data,a0	
	moveq.l		#0,d0
	move.w 		0x1c(a0),d0		/* Get loop offset */
	cmp.w 		#0,d0
	beq 		no_loop
	rol.w 		#8,d0			/* Convert to big-endian */
	add.w 		#0x1c,d0
	bra 		got_loop
	no_loop:
	move.w 		#0x40,d0		/* No loop offset - use the start of the song */
	got_loop:
	add.l		#music_data,d0
	move.l		d0,a5

	move.l		#music_data+0x40,a4	/* VGM data pointer in a4 */

	/* Allow the 68k to access the FM chip */
        move.w  	#0x100,0xa11100
	move.w  	#0x100,0xa11200

/* Main loop */
	
vgmloop:
#	bra vgmloop
read_cmd:
	move.b		#0x10,d7
	move.b		(a4)+,d0		/* Read one byte from the VGM data */

	cmp.b		#0x70,d0
	bmi		below_0x70
	cmp.b		#0x80,d0
	bmi		wait_n_short
	
	cmp.b		#0x90,d0
	bmi		pcm_write
	cmp.b		#0xe0,d0
	beq		pcm_seek
below_0x70:
	cmp.b		#0x4f,d0
	beq		pan
	cmp.b		#0x50,d0
	beq		psg_write
	cmp.b		#0x52,d0
	beq		ym2612_write_low
	cmp.b		#0x53,d0
	beq		ym2612_write_high
	cmp.b		#0x61,d0
	beq		wait_n
	cmp.b		#0x62,d0
	beq		wait_735
	cmp.b		#0x63,d0
	beq		wait_882
	cmp.b		#0x66,d0
	beq		data_end
	cmp.b		#0x67,d0
	beq		data_block

	bra		read_cmd	
pan:
	move.b		(a4)+,d0
	bra		read_cmd
psg_write:
	move.b		(a4)+,d0
	move.b		d0,0xc00011
	bra		read_cmd
ym2612_write_low:
	move.b		(a4)+,d0
	move.b		d0,0xa04000
	nop
	nop
	move.b		(a4)+,d1
	move.b		d1,0xa04001
	bra		read_cmd
ym2612_write_high:
	move.b		(a4)+,d0
	move.b		d0,0xa04002
	nop
	nop
	move.b		(a4)+,d0
	move.b		d0,0xa04003
	bra		read_cmd
wait_n_short:
	and.l		#15,d0
	addq.l		#1,d0
#	move.b		#0,d7
	jsr		delay_n_samples
	bra		read_cmd
wait_n:
	move.b		(a4)+,d0
	lsl.w		#8,d0
	move.b		(a4)+,d0
	rol.w		#8,d0
	jsr		delay_n_samples
	bra		read_cmd
wait_735:
	move.w		#735,d0
	jsr		delay_n_samples
	bra		read_cmd
wait_882:
	move.w		#882,d0
	jsr		delay_n_samples
	bra		read_cmd
pcm_write:
	move.b		(a3)+,d1
	move.b		#0x2a,0xa04000
	nop
	nop
	move.b		d1,0xa04001
	and.l		#15,d0
# lsr.l		#2,d0
	move.b		#0,d7
	jsr		delay_n_samples
	bra		read_cmd
pcm_seek:
	move.b		(a4)+,d0
	lsl.l		#8,d0
	move.b		(a4)+,d1
	or.b		d1,d0
	lsl.l		#8,d0
	move.b		(a4)+,d1
	or.b		d1,d0
	lsl.l		#8,d0
	move.b		(a4)+,d1
	or.b		d1,d0
	rol.w		#8,d0
	swap.w		d0
	rol.w		#8,d0
	move.l		a6,a3
	add.l		d0,a3
	bra		read_cmd
data_block:
	move.b		(a4)+,d0
	cmp.b		#0x66,d0
	bne		read_cmd
	move.b		(a4)+,d0
	cmp.b		#0x00,d0
	bne		read_cmd
	move.b		(a4)+,d0
	lsl.l		#8,d0

	move.b		(a4)+,d1
	or.b		d1,d0
	lsl.l		#8,d0
	move.b		(a4)+,d1
	or.b		d1,d0
	lsl.l		#8,d0
	move.b		(a4)+,d1
	or.b		d1,d0
	rol.w		#8,d0

	swap.w		d0
	rol.w		#8,d0
	move.l		a4,a6
	move.l		a6,a3
	move.l		d0,d6
	add.l		d0,a4
	bra		read_cmd	
data_end:
	move.l		a5,a4
	bra		read_cmd

	
delay_n_samples_short:
	cmp.w		#0,d0
	beq		delay_done
	move.l		d0,d1
	lsl.w		#5,d0
	sub.w		d1,d0
	lsr.w		#2,d0
	delay_loop:
		sub.l	#1,d0
		bne	delay_loop
	rts
	

/* Delay n/44100 s using YM2612 Timer A */
delay_n_samples:
	cmp.w		#0,d0
	beq		delay_done
	move.l		d0,d1
	lsl.l		#3,d0
	add.l		d1,d1
	add.l		d1,d0
	lsr.l		#3,d0
	move.l		d0,d2

	delay_again:
	move.l		d2,d0
	and.l		#0xFFF00,d0
	beq		low_left
	move.l		#256,d0
	bra		okdelay
	low_left:
	move.l		d2,d0
	and.w		#255,d0
	okdelay:
	eor.w		#0x3ff,d0
	addq.l		#1,d0
	move.l		d0,d1
	lsr		#2,d0
	and.w		#3,d1
	move.b		#0x24,0xa04000
	nop
	nop
	move.b		d0,0xa04001
	nop
	move.b		#0x25,0xa04000
	nop
	nop
	move.b		d1,0xa04001
	nop
	nop
	move.b		#0x27,0xa04000
	nop
	nop
	move.b		#0x15,0xa04001
	nop
	check_timer:
		move.b	0xa04000,d0
		and.b	#1,d0
		beq	check_timer
	cmp.b		#0,d7
	beq		delay_next
	move.b		#0x27,0xa04000
	nop
	nop
	move.b		d7,0xa04001
	delay_next:
	sub.l		#256,d2
	bpl		delay_again
	delay_done:
	rts
		
	
#################################################
#                                               #
#         Initialize VDP registers              #
#                                               #
#################################################

init_gfx:
	move.l 		#GFXCNTL,a3
	write_vdp_reg 	0,(VDP0_E_HBI + VDP0_E_DISPLAY + VDP0_PLTT_FULL)
	write_vdp_reg 	1,(VDP1_E_VBI + VDP1_E_DISPLAY + VDP1_E_DMA + VDP1_PAL + VDP1_RESERVED)
	write_vdp_reg 	2,(0xe000 >> 10)	/* Screen map A adress */
	write_vdp_reg 	3,(0xe000 >> 10)	/* Window address */
	write_vdp_reg 	4,(0xc000 >> 13)	/* Screen map B address */
	write_vdp_reg 	5,(0xfc00 >>  9)	/* Sprite address */
	write_vdp_reg 	6,0	
	write_vdp_reg	7,0			/* Border color */
	write_vdp_reg	8,1			/* Unused (?) */
	write_vdp_reg	9,1			/* Unused (?) */
	write_vdp_reg	10,1			/* Lines per hblank interrupt */
	write_vdp_reg	11,4			/* 2-cell vertical scrolling */
	write_vdp_reg	12,(VDP12_SCREEN_V224 + VDP12_SCREEN_H256 + VDP12_PROGRESSIVE)
	write_vdp_reg	13,(0x6000 >> 10)	/* Horizontal scroll address */
	write_vdp_reg	15,2
	write_vdp_reg	16,(VDP16_MAP_V32 + VDP16_MAP_H32)
	write_vdp_reg	17,0
	write_vdp_reg	18,0xff
	rts



#################################################
#                                               #
#        Put a string on screen map A           #
#                                               #
# Parameters:                                   #
#  a1: String pointer (ASCIIZ)                  # 
#  d0: X coordinate                             #
#  d1: Y coordinate                             #
#                                               #
#################################################

puts:
	lsl.l		#6,d1
	add.l		d0,d1
	add.l		d0,d1
	add.l		#0xe000,d1
	VRAM_ADDR_var	d0,d1
	move.l 		#GFXCNTL,a4
	move.l		d0,(a4)
	move.l 		#GFXDATA,a4
	moveq.l		#0,d0
_puts:
	move.b		(a1)+,d0
	beq		_puts_done
	sub.b		#32,d0
	move.w		d0,(a4)
	bra		_puts
_puts_done:
	rts

	

#################################################
#                                               #
#        Load tile data from ROM                #
#                                               #
# Parameters:                                   #
#  a3: VRAM base                                # 
#  a4: pattern address                          #
#  d4: number of tiles to load                  #
#                                               #
#################################################

load_tiles:
	move.l 		#GFXCNTL,a2
	VRAM_ADDR_var 	d0,a3
	move.l 		d0,(a2)
	lsl		#3,d4
	
	move.l 		#GFXDATA,a3
	subq.w 		#1,d4		/* DBRA stops at -1, so subtract 1 from the counter first */
_copy_tile_data:
	move.l 		(a4)+,(a3)
	dbra 		d4,_copy_tile_data
	rts


#################################################
#                                               #
#        Clear one of the screen maps           #
#                                               #
# Parameters:                                   #
#  a0: Map address                              # 
#  d0: Data to write to each map entry          #
#                                               #
#################################################

clear_map:
	move.l 		#GFXCNTL,a4
	VRAM_ADDR_var	d1,a0
	move.l 		d1,(a4)
	move.l 		#GFXDATA,a3
	move.w		#1023,d1	/* Loop counter */
_clear_map_loop:
	move.w		d0,(a3)
	move.w		d0,(a3)
	dbra		d1,_clear_map_loop
	rts
	

#################################################
#                                               #
#        Load color data from ROM               #
#                                               #
# Parameters:                                   #
#  a3: CRAM base                                # 
#  a4: color list address                       #
#  d4: number of colors to load                 #
#                                               #
#################################################

load_colors:
	move.l 		#GFXCNTL,a2
	CRAM_ADDR_var 	d0,a3
	move.l 		d0,(a2)

	move.l 		#GFXDATA,a3
	subq.w		#1,d4
_copy_color_data:
	move.w		(a4)+,(a3)
	dbra		d4,_copy_color_data

	rts


	
#################################################
#                                               #
#       Wait for next VBlank interrupt          #
#                                               #
#################################################

wait_vsync:
	movea.l		#vtimer,a0
	move.l		(a0),a1
_wait_change:
	cmp.l		(a0),a1
	beq		_wait_change
	rts


#################################################
#                                               #
#                 ROM DATA                      #
#                                               #
#################################################


.align 4
palette:
	dc.w 0x0203,0x0ecc
	

.equ music_data, palette+24

.end






